@@ -1,57 +1,120 @@
 # import {{{1
 import json
-from django.db.models.functions import Concat
-from django.core import serializers
-from django.urls import path
-from django.http import HttpResponse, FileResponse
-from django.shortcuts import render, redirect
-from django.views.generic.base import View, RedirectView
-from django.views.generic.edit import FormView
-from django.views.generic import DetailView, CreateView, UpdateView, DeleteView, ListView
-from django.utils.translation import gettext_lazy as _
-from django.urls import reverse_lazy
-from django.contrib.auth.mixins import PermissionRequiredMixin
-
 import pyexcel
 
-from .forms import get_generic_form, ImportXlsForm
-from .core import get_color_from_class, get_layout_data, get_model_fields_raw, get_model_fields
+from django.http import JsonResponse
+from django.conf import settings
+from django.urls import path
+from django.http import FileResponse
+from django.urls import reverse_lazy
+from django.shortcuts import render, redirect
 
-# list {{{1
-class GenericListView(PermissionRequiredMixin, ListView):
-    template_name = 'genviews/list.html'
-    list_cls = 'generic-list'
-    fields = "__all__" 
+from django.views.generic import DetailView, CreateView, UpdateView, DeleteView, ListView
+from django.views.generic.base import View, TemplateView
+from django.views.generic.edit import FormView
+from django.contrib.auth.mixins import PermissionRequiredMixin
+from django.contrib.auth import views as auth_views
+
+from django.utils.translation import gettext_lazy as _
+
+from .forms import get_generic_form, ImportXlsForm
+from .core import get_color_from_class, get_layout_data, get_model_fields
+from .models import GenericModel
+
+# LOGIN {{{1
+class GenericLoginView(auth_views.LoginView):
+    extra_context = {
+        'settings': settings,
+        'test': 'test'
+    }
+
+# DASHBOARD {{{1
+class GenericDashboardView(PermissionRequiredMixin, TemplateView):
+    models = []
+    template_name = 'genviews/dashboard.html'
+    template_list_name = 'genviews/dashboard_list.html'
+    list_cls = 'dashboard-list'
+    auth = 'superuser'
 
     def __init__(self, **kwargs):
-        self.current_app = self.model._meta.app_label
-        self.permission_required = f'{self.current_app}.view_{self.model.__name__.lower()}'
+        assert hasattr(settings, 'DASH_LIST_COUNT'), f"DASH_LIST_COUNT non défini dans settings.py, il faut définir la valeur par default du nombre d'element à afficher dans le tableau de bord"
+        self.permission_required = f'{self.auth}.dashboard'
         super().__init__(**kwargs)
 
     def get_context_data(self, **kwargs):
         context = super().get_context_data(**kwargs)
-        context['list_title'] = self.get_list_title();
-        context['fields'] = self.get_fields();
-        context['list_cls'] = self.list_cls
-        context['list_menu'] = self.model.get_list_menu()
+        context['models'] = {}
+        for m in self.models:
+            model_dict = context['models'][m] = {}
+            obj_count = m.get_view_data('dashboard', 'count') or settings.DASH_LIST_COUNT
+            model_dict['title'] = m.get_view_data('dashboard', 'title')
+            model_dict['html'] = m.get_view_data('dashboard', 'html') or self.template_list_name
+            model_dict['menu'] = m.get_dash_menu()
+            model_dict['fields'] = self.get_dash_fields(m)
+            model_dict['list'] = m.objects.all()[:obj_count]
+
+        context['settings'] = settings
         context['layout_data'] = get_layout_data()
         return context
 
-    def get_list_title(self):
-        if hasattr(self.model, "LIST_TITLE"):
-            return self.model.LIST_TITLE
-        else:
-            return self.model._meta.verbose_name_plural;
+    def get_dash_fields(self, model):
+        view_struct_fields = model.get_view_data('dashboard')
+        return get_model_fields(model, view_struct_fields)
 
-    def get_fields(self):
-        if not hasattr(self.model, "VIEWS_STRUCT"):
-            print(f"Warning : trying to generate generic list without VIEW_STRUCT with model {self.model.__name__}")
-            return get_model_fields(self.model, '__all__')
+# OMNIBAR {{{1
+# class GenericOmnibarView(PermissionRequiredMixin, View):
+class GenericOmnibarView(View):
+    models = []
+    auth = 'superuser'
 
-        fields = self.model.VIEWS_STRUCT["list"]["fields"]
-        return get_model_fields(self.model, fields)
+    def __init__(self, **kwargs):
+        self.permission_required = f'{self.auth}.omnibar'
+        super().__init__(**kwargs)
 
-# create {{{1
+    def get(self, request, *args, **kwargs):
+        d = [] 
+        
+        for m in self.models:
+            assert issubclass(m, GenericModel), f'{m.get_id()} Must inherit from GenericModel' 
+            for o in m.objects.all():
+                d.append({
+                    'type': m.get_id(),
+                    'handler': 'toDetail',
+                    'param0': o.name,
+                    'param1': o.get_absolute_url()
+                })
+        return JsonResponse(d, safe=False)
+        
+# LIST {{{1
+class GenericListView(PermissionRequiredMixin, ListView):
+    template_name = 'genviews/list.html'
+    list_cls = 'generic-list'
+    fields = '__all__' 
+
+    def __init__(self, **kwargs):
+        assert hasattr(settings, 'FIELDMAP'), f"Pas de template FIELDMAP défini dans settings.py, il faut définir comment chaque field de la liste va s'afficher dans un template"
+        self.current_app = self.model._meta.app_label
+        self.permission_required = f'{self.current_app}.list_{self.model.__name__.lower()}'
+        super().__init__(**kwargs)
+
+    def get_context_data(self, **kwargs):
+        context = super().get_context_data(**kwargs)
+
+        context['list_title'] = self.model.get_view_data('list', 'title')
+        context['list_menu'] = self.model.get_list_menu()
+        context['list_cls'] = self.list_cls
+        context['fields'] = self.get_list_fields();
+
+        context['settings'] = settings
+        context['layout_data'] = get_layout_data()
+        return context
+
+    def get_list_fields(self):
+        view_struct_fields = self.model.get_view_data('list')
+        return get_model_fields(self.model, view_struct_fields)
+
+
+# CREATE {{{1
 class GenericCreateView(PermissionRequiredMixin, CreateView):
     template_name = 'genviews/form.html'
     color_classes = []
@@ -61,39 +124,43 @@ class GenericCreateView(PermissionRequiredMixin, CreateView):
         self.permission_required = f'{self.current_app}.add_{self.model.__name__.lower()}'
 
         if not self.form_class:
- 
-            if hasattr(self.model, 'VIEWS_STRUCT'):
-                if 'create' in self.model.VIEWS_STRUCT and 'fields' in self.model.VIEWS_STRUCT['create']:
-                    self.fields = self.model.VIEWS_STRUCT['create']['fields']
-
-            self.form_class = get_generic_form(self.model, self.fields)
-            self.fields = None
+            self.form_class = self.create_generic_form()
             self.initial = {'model': self.model}
+            # both fields and form_class not permitted
+            self.fields = None
             super().__init__(**kwargs)
 
-    def get_context_data(self, **kwargs):
-        current_app = self.request.resolver_match.namespace 
-        context = super().get_context_data(**kwargs)
-        context['form_title'] = _("Nouveau") + " " + \
-              self.model._meta.verbose_name.lower()
-        context['form_action'] = reverse_lazy(f"{current_app}:{self.model.__name__.lower()}-create")
-        context['layout_data'] = get_layout_data()
+    def create_generic_form(self):
+        view_struct_fields = self.model.get_view_data('create')
+        return get_generic_form(self.model, view_struct_fields) 
 
+    def get_context_data(self, **kwargs):
+        if 'param' in self.kwargs and 'pk' in self.kwargs:
+            self.initial = {
+                self.kwargs['param']: self.kwargs['pk']
+            }
+
+        context = super().get_context_data(**kwargs)
+
+        context['form_title'] = self.model.get_view_data('create', 'title')
+        context['form_action'] = reverse_lazy(f"{self.current_app}:{self.model.__name__.lower()}-create")
         if len(self.color_classes) > 0:
             context['color_palette'] = get_color_from_class(self.color_classes)
-       
-
-        print(context)
 
+        context['settings'] = settings
+        context['layout_data'] = get_layout_data()
         return context
 
     def get_success_url(self):
         if self.success_url:
             return reverse_lazy(self.success_url)
-        else:
-            current_app = self.request.resolver_match.namespace
-            return reverse_lazy(f"{current_app}:{self.model.__name__.lower()}-list")
-#update {{{1
+
+        for v, args in (('detail', [self.object.pk]), ('list', None), ('create', None)):
+            if self.model.get_view_data(v):
+                url = f"{self.current_app}:{self.model.__name__.lower()}-{v}"
+                return reverse_lazy(url, args=args)
+
+# UPDATE {{{1
 class GenericUpdateView(PermissionRequiredMixin, UpdateView):
     template_name = 'genviews/form.html'
     color_classes = []
@@ -103,80 +170,67 @@ class GenericUpdateView(PermissionRequiredMixin, UpdateView):
         self.permission_required = f'{self.current_app}.change_{self.model.__name__.lower()}'
 
         if not self.form_class:
-            self.form_class = get_generic_form(self.model, self.fields)
+            self.form_class = self.create_generic_form()
+            # both fields and form_class not permitted
             self.fields = None
             self.initial = {'model': self.model}
             super().__init__(**kwargs)
 
+    def create_generic_form(self):
+        view_struct_fields = self.model.get_view_data('update')
+        return get_generic_form(self.model, view_struct_fields) 
+
     def get_context_data(self, **kwargs):
-        current_app = self.request.resolver_match.namespace
         context = super().get_context_data(**kwargs)
-        context['form_title'] = f"{self.model._meta.verbose_name} {str(self.object)}"
+
+        context['form_title'] = self.model.get_view_data('update', 'title')
         context['form_action'] = reverse_lazy(
-          f"{current_app}:{self.model.__name__.lower()}-update", args=[self.object.pk])
-        context['layout_data'] = get_layout_data()
+          f"{self.current_app}:{self.model.__name__.lower()}-update", args=[self.object.pk])
         
         if len(self.color_classes) > 0:
             context['color_palette'] = get_color_from_class(self.color_classes)
+
+        context['settings'] = settings
+        context['layout_data'] = get_layout_data()
         return context
 
     def get_success_url(self):
         if self.success_url:
             return reverse_lazy(self.success_url)
-        else:
-            current_app = self.model._meta.app_label
-            return reverse_lazy(f"{current_app}:{self.model.__name__.lower()}-detail", args=[self.object.pk])
 
-# detail {{{1
+        for v, args in (('detail', [self.object.pk]), ('list', None), ('create', None)):
+            if self.model.get_view_data(v):
+                url = f"{self.current_app}:{self.model.__name__.lower()}-{v}"
+                return reverse_lazy(url, args=args)
+
+# DETAIL {{{1
 class GenericDetailView(PermissionRequiredMixin, DetailView):
     template_name = 'genviews/detail.html'
 
     def __init__(self, **kwargs):
         self.current_app = self.model._meta.app_label
-        self.permission_required = f'{self.current_app}.view_{self.model.__name__.lower()}'
+        self.permission_required = f'{self.current_app}.detail_{self.model.__name__.lower()}'
         super().__init__(**kwargs)
 
     def get_context_data(self, **kwargs):
-        current_app = self.request.resolver_match.namespace
-
         context = super().get_context_data(**kwargs)
-        context['layout_data'] = get_layout_data()
-        
-        if hasattr(self.model, 'VIEWS_STRUCT'):
-            if 'detail' in self.model.VIEWS_STRUCT and 'tpl' in self.model.VIEWS_STRUCT['detail']:
-                context['obj_tpl'] = self.model.VIEWS_STRUCT['detail']['tpl']
 
-        if hasattr(self, 'actions_extra'):
-            context['actions_extra'] = self.actions_extra
+        context['obj_html'] = self.model.get_view_data('detail', 'detail_html')
+        context['MEDIA_URL'] = settings.MEDIA_URL
+        context['detail_menu'] = self.object.get_view_menu('detail')
 
-        if hasattr(self, 'template_extra'):
-            context['template_extra'] = self.template_extra
+        # TODO utile ??
+        #if hasattr(self, 'actions_extra'):
+        #    context['actions_extra'] = self.actions_extra
 
-        return context
-#delete {{{1
-class GenericDeleteView(PermissionRequiredMixin, DeleteView):
-    template_name = 'genviews/delete.html'
+        #if hasattr(self, 'template_extra'):
+        #    context['template_extra'] = self.template_extra
 
-    def __init__(self, **kwargs):
-        self.current_app = self.model._meta.app_label
-        self.permission_required = f'{self.current_app}.delete_{self.model.__name__.lower()}'
-        super().__init__(**kwargs)
-
-    def get_context_data(self, **kwargs):
-        current_app = self.request.resolver_match.namespace
-        context = super().get_context_data(**kwargs)
-        context['back_url'] = f"{current_app}:{self.model.__name__.lower()}-detail"
+        context['settings'] = settings
         context['layout_data'] = get_layout_data()
         return context
 
-    def get_success_url(self):
-        if self.success_url:
-            return reverse_lazy(self.success_url)
-        else:
-            current_app = self.request.resolver_match.namespace
-            return reverse_lazy(f"{current_app}:{self.model.__name__.lower()}-list")
-
-#history {{{1
+# HISTORY {{{1
 class GenericHistoryView(PermissionRequiredMixin, DetailView):
     template_name = 'genviews/history.html'
 
@@ -186,14 +240,41 @@ class GenericHistoryView(PermissionRequiredMixin, DetailView):
         super().__init__(**kwargs)
 
     def get_context_data(self, **kwargs):
-        current_app = self.request.resolver_match.namespace
         context = super().get_context_data(**kwargs)
-        context['layout_data'] = get_layout_data()
+        context['list_title'] = self.model.get_view_data('history', 'title')
+        context['list_menu'] = self.object.get_view_menu('history')
         context['object_list'] = self.object.history.all()
         context['fields'] = get_model_fields(self.model, self.fields)
+
+        context['settings'] = settings
+        context['layout_data'] = get_layout_data()
         return context
 
-#exportxls {{{1
+
+# DELETE {{{1
+class GenericDeleteView(PermissionRequiredMixin, DeleteView):
+    template_name = 'genviews/delete.html'
+
+    def __init__(self, **kwargs):
+        self.current_app = self.model._meta.app_label
+        self.permission_required = f'{self.current_app}.delete_{self.model.__name__.lower()}'
+        super().__init__(**kwargs)
+
+    def get_context_data(self, **kwargs):
+        context = super().get_context_data(**kwargs)
+        context['back_url'] = f"{self.current_app}:{self.model.__name__.lower()}-detail"
+
+        context['settings'] = settings
+        context['layout_data'] = get_layout_data()
+        return context
+
+    def get_success_url(self):
+        for v, args in (('list', None), ('create', None)):
+            if self.model.get_view_data(v):
+                url = f"{self.current_app}:{self.model.__name__.lower()}-{v}"
+                return reverse_lazy(url, args=args)
+
+# EXPORT XLS {{{1
 class GenericExportxlsView(PermissionRequiredMixin, View):
     def __init__(self, **kwargs):
         self.current_app = self.model._meta.app_label
@@ -203,23 +284,26 @@ class GenericExportxlsView(PermissionRequiredMixin, View):
     def get(self, request, *args, **kwargs):
         if hasattr(self.model, 'get_excel_dict'):
             xls_file = pyexcel.save_as(dest_file_type="xls", adict=self.model.get_excel_dict())
-            response = FileResponse(xls_file, as_attachment=True,
+            return FileResponse(xls_file, as_attachment=True,
                                     filename=f"{self.model.__name__.lower()}_export.xls")
-            return response
-
         else: 
-            if self.fields == '__all__':
+            tuple_fields = self.get_exportxls_fields();
+            tuple_fields.reverse()
+            if tuple_fields == '__all__':
                 objects = self.model.objects.values()
             else:
-                # objects = self.model.objects.values(*self.fields)
-                objects = self.model.objects.values(*self.fields)
+                objects = self.model.objects.values(*tuple_fields)
 
             xls_file = pyexcel.save_as(dest_file_type="xls", records=objects)
-            response = FileResponse(xls_file, as_attachment=True,
+            return FileResponse(xls_file, as_attachment=True,
                   filename=f"{self.model.__name__.lower()}_export.xls")
-            return response
 
-#importxls {{{1
+    def get_exportxls_fields(self):
+        view_struct_fields = self.model.get_view_data('exportxls')
+        return view_struct_fields
+
+
+# IMPORT XLS {{{1
 class GenericImportxlsView(PermissionRequiredMixin, FormView):
     template_name = 'genviews/import_xls.html'
     form_class = ImportXlsForm
@@ -232,15 +316,20 @@ class GenericImportxlsView(PermissionRequiredMixin, FormView):
 
 #get context data {{{2
     def get_context_data(self, **kwargs):
-        current_app = self.request.resolver_match.namespace
         context = super().get_context_data(**kwargs)
         context['form_title'] = f"Importer des données excel"
         context['form_subtitle'] = self.model._meta.verbose_name
-        context['fields'] = self.model.get_fields_data(self.fields)
+        context['fields'] = self.get_import_fields()
         context['import_sample'] = self.model.get_samples()
         context['layout_data'] = get_layout_data()
+        context['settings'] = settings
         return context
 
+    def get_import_fields(self):
+        view_struct_fields = self.model.get_view_data('importxls')
+        return get_model_fields(self.model, view_struct_fields)
+
+
 #form_valid {{{2
     def form_valid(self, form):
         xls_file = form.files['xls_file']
@@ -258,7 +347,8 @@ class GenericImportxlsView(PermissionRequiredMixin, FormView):
         return render(self.request, self.template_name,
               context=context)
 
-      #importxls save {{{1
+
+#IMPORT XLS SAVE {{{1
 class GenericImportxlssaveView(PermissionRequiredMixin, View):
     def post(self, request):
         for key, value in request.POST.items():
@@ -279,9 +369,6 @@ class GenericImportxlssaveView(PermissionRequiredMixin, View):
         self.permission_required = f'{self.current_app}.importxlssave_{self.model.__name__.lower()}'
         super().__init__(**kwargs)
 
-
-
-
 #   def form_valid(self, form):
 #      def testinit(row):
 #         # self.model.book_category.field.related_model
@@ -332,24 +419,30 @@ class GenericImportxlssaveView(PermissionRequiredMixin, View):
 #   pass
 
 # create_model_paths {{{1
-def get_gen_path_from_model(model):
+def create_model_paths(model):
     mod_name = model.__name__
     mod_name_low = mod_name.lower()
     path_to_return = []
 # get_fields {{{
     def get_fields(view):
-        fields = model.get_view_fields(view.lower())
-        if fields[0] == "__all__":
-            return get_model_fields_raw(model, "__all__")
+        fields = model.get_view_data(view.lower())
+        if fields == '__all__':
+            return get_model_fields(model)
         else:
             return fields
+        if view == 'Exportxls':
+            print(fields)
+        return get_model_fields(model, fields)
 # }}}
-    
-    if not hasattr(model, "GENERIC_VIEW_LIST"):
-        print(f"Warning : trying to generate generic views without GENERIC_VIEW_LIST with model {model.__name__}")
-        return []
 
-    for view in model.GENERIC_VIEW_LIST:
+    assert issubclass(model, GenericModel), 'Not herited from GenericModel'
+    assert hasattr(model, 'get_views_struct'), 'Not herited from GenericModel'
+
+    for view in model.get_views_struct().keys():
+        view = view.title()
+
+        assert globals().get(f"Generic" + view + "View"), f"Error : trying to set generic views Generic{view}View for model {model}"
+
         # print(get_fields(view))
         # print(mod_name + view)
         ClassView = type(
@@ -375,4 +468,15 @@ def get_gen_path_from_model(model):
                   name=mod_name_low + '-' + view.lower()
                   )
               )
+
+        if view == 'Create': 
+            view_path += '/<str:param>/<int:pk>'
+            path_to_return.append(
+                  path(
+                      view_path,
+                      ClassView.as_view(),
+                      name=mod_name_low + '-' + view.lower()
+                      )
+                  )
+ 
     return path_to_return
